



/*
 *
 * Multithreaded HTTP Download Accelarator: Aget
 * (c) 2002, Murat Balaban <murat@enderunix.org>
 *
 * See COPYING for copyright and copying restrictions
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>


#include <netdb.h>

#include <sys/socket.h>
#include <sys/types.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include "socket.h"

extern int errno;


void head_req(struct request *req)
{
	struct sockaddr_in sin;
	struct hostent *he;
	int sd;
	char *sbuf;
	char *rbuf;
	char *tok;
	char *s;
	int clength;


	sbuf = (char *)calloc(HEADREQSIZ + strlen(req->url), sizeof(char));
	rbuf = (char *)calloc(HEADRECVSIZ, sizeof(char));

	if ((he = gethostbyname(req->host)) == NULL) {
		fprintf(stderr, "Error: Cannot resolve hostname\n");
		herror("gethostbyname");
		exit(1);
	}
	strncpy(req->ip, inet_ntoa(*(struct in_addr *)he->h_addr), MAXIPSIZ);

	bzero(&sin, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr =inet_addr(req->ip);
	sin.sin_port = htons(req->port);

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		exit(1);
	}
	if ((connect(sd, (const struct sockaddr *)&sin, sizeof(sin))) == -1) {
		perror("connect");
		exit(1);
	}
	Log("Head-Request Connection established");

	sprintf(sbuf, HEADREQ, req->url, req->host, PROGVERSION);
	if ((send(sd, sbuf, strlen(sbuf), 0)) == -1) {
		perror("send");
		exit(1);
	}

	if ((recv(sd, rbuf, HEADRECVSIZ, 0)) == -1) {
		perror("recv");
		exit(1);
	}

        if ((strstr((tok = strtok(rbuf, "\r\n")), "HTTP/1.1 404")) != NULL) {
                printf("<LOG> File Not Found!, Aborting...\n");
	                close(sd);
	                exit(1);
        } else
        if ((strstr(tok, "HTTP/1.1 200")) != NULL) {
                while ((tok = strtok(NULL, "\r\n")) != NULL) {
                        if ((strstr(tok, "Content-Length")) != NULL) {
                                s = (tok + strlen("Content-Length: "));
                                clength = atoi(s);
                                printf("<LOG> Content-Length: %d\n", clength);
				req->clength = clength;
                        }
                }
        }

	free(sbuf);
	free(rbuf);

}



void *partialGetHandler(void *arg) {
	thread_data *td;
	struct sockaddr_in sin;
	int sd;
	char *req, *rbuf, *s;
	int offset, dr, dw, i; 
	long foffset;
	pthread_t tid;
	FILE *fp;
	char *file;

	tid = pthread_self();
	
        printf("<THREAD #%ld> Entering thread ...\n", tid); 

	td = (thread_data *)arg;
	file = (char *)calloc(strlen(td->file) + 10, sizeof(char));
	sprintf(file, "%s.tmp%d", td->file, td->tind);

	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "<THREAD #%ld> cannot create temporary file %s: %s\n", tid, file, strerror(errno));
		exit(1);
	}
	foffset = td->foffset;
        printf("<THREAD #%ld> Range: %ld-%ld\n", tid, td->soffset, td->foffset); 

	bzero(&sin, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addr(td->ip);
	sin.sin_port = htons(td->port);

	req = (char *)calloc(GETREQSIZ, sizeof(char));
	rbuf = (char *)calloc(GETRECVSIZ, sizeof(char));
	sprintf(req, GETREQ, td->url, td->host, PROGVERSION, td->soffset);

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("thread socket");
		free(td);
		free(req);
		pthread_exit((void *)1);
	}
	if ((connect(sd, (const struct sockaddr *)&sin, sizeof(sin))) == -1) {
		perror("thread connect");
		free(td);
		free(req);
		pthread_exit((void *)1);
	}

	if ((send(sd, req, strlen(req), 0)) == -1) {
		perror("send");
		free(td);
		free(req);
		pthread_exit((void *)1);
	}

        if ((dr = recv(sd, rbuf, HEADRECVSIZ, 0)) == -1) {
                perror("recv");
		free(td);
		free(req);
		pthread_exit((void *)1);
        }

        if ((strstr(rbuf, "HTTP/1.1 404")) != NULL) {
                printf("<THREAD #%ld> Server returned HTTP/1.1 404 - File Not Found, aborting\n", tid);
		free(td);
		free(req);
                close(sd);
                exit(1);
        } else
        if ((strstr(rbuf, "HTTP/1.1 206")) != NULL) 
                printf("<THREAD #%ld> HTTP/1.1 206 - Partial Content\n", tid);

	s = rbuf;
	i = 0;
	while(1) {
		if (*s == '\n' && *(s - 1) == '\r' && *(s - 2) == '\n' && *(s - 3) == '\r') {
			s++;
			i++;
			break;
		}
		s++;
		i++;
	}

	if ((dr - i ) > foffset) 
		dw = fwrite(s, sizeof(char), (foffset - i), fp);
	else
		dw = fwrite(s, sizeof(char), (dr - i), fp);
	offset = td->soffset + dw;

	while (offset < foffset) {
		memset(rbuf, GETRECVSIZ, 0);
		dr = recv(sd, rbuf, GETRECVSIZ, 0);
		if ((offset + dr) > foffset)
			dw = fwrite(rbuf, sizeof(char), foffset - offset, fp);
		else
			dw = fwrite(rbuf, sizeof(char), dr, fp);
		offset += dw;
        	/*printf("<THREAD #%d> offset: %d bytes, foffset: %d\n", tid, offset, foffset);*/
	}

	free(req);
	free(td);
	close(sd);
	fclose(fp);


        printf("<THREAD #%ld> Leaving thread...\n", tid);

	pthread_mutex_lock(&thread_exit_mtx);
	thread_exit_cnt--;
	pthread_cond_signal(&thread_exit_cnd);
	pthread_mutex_unlock(&thread_exit_mtx);

	pthread_exit(NULL);
	return NULL;
}


void Log(char *log)
{
#ifdef DEBUG
	fprintf(stderr, "<LOG> %s\n", log);
#endif
}
